import copy
import random
import time

import numpy as np

from pymoo.core.callback import Callback
from pymoo.core.evaluator import Evaluator
from pymoo.core.population import Population
from pymoo.core.result import Result
from pymoo.util.function_loader import FunctionLoader
from pymoo.util.misc import termination_from_tuple
from pymoo.util.optimum import filter_optimum


class Algorithm:
    """

    This class represents the abstract class for any algorithm to be implemented. Most importantly it
    provides the solve method that is used to optimize a given problem.

    The solve method provides a wrapper function which does validate the input.


    Parameters
    ----------

    problem : :class:`~pymoo.core.problem.Problem`
        Problem to be solved by the algorithm

    termination: :class:`~pymoo.core.termination.Termination`
        Object that tells the algorithm when to terminate.

    seed : int
        Random seed to be used. Same seed is supposed to return the same result. If set to None, a random seed
        is chosen randomly and stored in the result object to ensure reproducibility.

    verbose : bool
        If true information during the algorithm execution are displayed

    callback : func
        A callback function can be passed that is executed every generation. The parameters for the function
        are the algorithm itself, the number of evaluations so far and the current population.

            def callback(algorithm):
                pass

    save_history : bool
        If true, a current snapshot of each generation is saved.

    pf : numpy.array
        The Pareto-front for the given problem. If provided performance metrics are printed during execution.

    return_least_infeasible : bool
        Whether the algorithm should return the least infeasible solution, if no solution was found.

    evaluator : :class:`~pymoo.core.evaluator.Evaluator`
        The evaluator which can be used to make modifications before calling the evaluate function of a problem.


    """

    def __init__(self, **kwargs):

        # !
        # Here all algorithm parameters needed no matter what is problem is passed are defined
        # problem dependent initialization happens in setup(problem, **kwargs)
        # !

        super().__init__()

        # prints the compile warning if enabled
        FunctionLoader.get_instance()

        # !
        # DEFAULT SETTINGS OF ALGORITHM
        # !
        # the termination criterion to be used by the algorithm - might be specific for an algorithm
        self.termination = kwargs.get("termination")
        # set the display variable supplied to the algorithm - might be specific for an algorithm
        self.display = kwargs.get("display")
        # callback to be executed each generation
        self.callback = kwargs.get("callback")

        # !
        # Attributes to be set later on for each problem run
        # !
        # the optimization problem as an instance
        self.problem = None
        # whether the algorithm should finally return the least infeasible solution if no feasible found
        self.return_least_infeasible = None
        # whether the history should be saved or not
        self.save_history = None
        # whether the algorithm should print output in this run or not
        self.verbose = None
        # the random seed that was used
        self.seed = None
        # an algorithm can defined the default termination which can be overwritten
        self.default_termination = None
        # whether the algorithm as terminated or not
        self.has_terminated = None
        # the pareto-front of the problem - if it exist or passed
        self.pf = None
        # the function evaluator object (can be used to inject code)
        self.evaluator = None
        # the current number of generation or iteration
        self.n_gen = None
        # the history object which contains the list
        self.history = None
        # the current solutions stored - here considered as population
        self.pop = None
        # a placeholder object for implementation to store solutions in each iteration
        self.off = None
        # the optimum found by the algorithm
        self.opt = None
        # can be used to store additional data in submodules
        self.data = {}
        # if the initialized method has been called before or not
        self.is_initialized = False
        # the time when the algorithm has been setup for the first time
        self.start_time = None

    # =========================================================================================================
    # PUBLIC
    # =========================================================================================================

    def setup(self,
              problem,

              # START Overwrite by minimize
              termination=None,
              callback=None,
              display=None,
              # END Overwrite by minimize

              # START Default minimize
              seed=None,
              verbose=False,
              save_history=False,
              return_least_infeasible=False,
              # END Default minimize

              pf=True,
              evaluator=None,
              **kwargs):

        # set the problem that is optimized for the current run
        self.problem = problem

        # set the provided pareto front
        self.pf = pf

        # by default make sure an evaluator exists if nothing is passed
        if evaluator is None:
            evaluator = Evaluator()
        self.evaluator = evaluator

        # !
        # START Default minimize
        # !
        # if this run should be verbose or not
        self.verbose = verbose
        # whether the least infeasible should be returned or not
        self.return_least_infeasible = return_least_infeasible
        # whether the history should be stored or not
        self.save_history = save_history

        # set the random seed in the algorithm object
        self.seed = seed
        if self.seed is None:
            self.seed = np.random.randint(0, 10000000)
        # set the random seed for Python and Numpy methods
        random.seed(self.seed)
        np.random.seed(self.seed)
        # !
        # END Default minimize
        # !

        # !
        # START Overwrite by minimize
        # !

        # the termination criterion to be used to stop the algorithm
        if self.termination is None:
            self.termination = termination_from_tuple(termination)
        # if nothing given fall back to default
        if self.termination is None:
            self.termination = self.default_termination

        if callback is not None:
            self.callback = callback

        if display is not None:
            self.display = display

        # !
        # END Overwrite by minimize
        # !

        # no call the algorithm specific setup given the problem
        self._setup(problem, **kwargs)

        return self

    def run(self):

        # now the termination criterion should be set
        if self.termination is None:
            raise Exception("No termination criterion defined and algorithm has no default termination implemented!")

        # while termination criterion not fulfilled
        while self.has_next():
            self.next()

        # create the result object to be returned
        res = self.result()

        return res

    def has_next(self):
        return not self.has_terminated

    def finalize(self):
        return self._finalize()

    def next(self):

        # get the infill solutions
        infills = self.infill()

        # call the advance with them after evaluation
        if infills is not None:
            self.evaluator.eval(self.problem, infills, algorithm=self)
            self.advance(infills=infills)

        # if the algorithm does not follow the infill-advance scheme just call advance
        else:
            self.advance()

    def _initialize(self):

        # the time starts whenever this method is called
        self.start_time = time.time()

        # set the attribute for the optimization method to start
        self.n_gen = 1
        self.has_terminated = False
        self.pop, self.opt = Population(), None

        # if the history is supposed to be saved
        if self.save_history:
            self.history = []

    def infill(self):
        if self.problem is None:
            raise Exception("Please call `setup(problem)` before calling next().")

        # the first time next is called simply initial the algorithm - makes the interface cleaner
        if not self.is_initialized:

            # hook mostly used by the class to happen before even to initialize
            self._initialize()

            # execute the initialization infill of the algorithm
            infills = self._initialize_infill()

        else:
            # request the infill solutions if the algorithm has implemented it
            infills = self._infill()

        # set the current generation to the offsprings
        if infills is not None:
            infills.set("n_gen", self.n_gen)

        return infills

    def advance(self, infills=None, **kwargs):

        # if infills have been provided set them as offsprings and feed them into advance
        self.off = infills

        # if the algorithm has not been already initialized
        if not self.is_initialized:

            # set the generation counter to 1
            self.n_gen = 1

            # assign the population to the algorithm
            self.pop = infills

            # do whats necessary after the initialization
            self._initialize_advance(infills=infills, **kwargs)

            # set this algorithm to be initialized
            self.is_initialized = True

        else:

            # increase the generation counter by one
            self.n_gen += 1

            # call the implementation of the advance method - if the infill is not None
            self._advance(infills=infills, **kwargs)

        # execute everything which needs to be done after having the algorithm advanced to the next generation
        self._post_advance()

        # set whether the algorithm has terminated or not
        self.has_terminated = not self.termination.do_continue(self)

        # if the algorithm has terminated call the finalize method
        if self.has_terminated:
            return self.finalize()

    def result(self):
        res = Result()

        # store the time when the algorithm as finished
        res.start_time = self.start_time
        res.end_time = time.time()
        res.exec_time = res.end_time - res.start_time

        res.pop = self.pop

        # get the optimal solution found
        opt = self.opt
        if opt is None or len(opt) == 0:
            opt = None

        # if no feasible solution has been found
        elif not np.any(opt.get("feasible")):
            if self.return_least_infeasible:
                opt = filter_optimum(opt, least_infeasible=True)
            else:
                opt = None
        res.opt = opt

        # if optimum is set to none to not report anything
        if res.opt is None:
            X, F, CV, G = None, None, None, None

        # otherwise get the values from the population
        else:
            X, F, CV, G = self.opt.get("X", "F", "CV", "G")

            # if single-objective problem and only one solution was found - create a 1d array
            if self.problem.n_obj == 1 and len(X) == 1:
                X, F, CV, G = X[0], F[0], CV[0], G[0]

        # set all the individual values
        res.X, res.F, res.CV, res.G = X, F, CV, G

        # create the result object
        res.problem, res.pf = self.problem, self.pf
        res.history = self.history

        return res

    def ask(self, *args, **kwargs):
        return self.infill(*args, **kwargs)

    def tell(self, *args, **kwargs):
        return self.advance(*args, **kwargs)

    # =========================================================================================================
    # PROTECTED
    # =========================================================================================================

    def _set_optimum(self):
        self.opt = filter_optimum(self.pop, least_infeasible=True)

    def _post_advance(self):

        # update the current optimum of the algorithm
        self._set_optimum()

        # display the output if defined by the algorithm
        if self.verbose and self.display is not None:
            self.display.do(self.problem, self.evaluator, self, pf=self.pf)

        # if a callback function is provided it is called after each iteration
        if self.callback is not None:
            if isinstance(self.callback, Callback):
                self.callback.notify(self)
            else:
                self.callback(self)

        if self.save_history:
            _hist, _callback = self.history, self.callback

            self.history, self.callback = None, None
            obj = copy.deepcopy(self)

            self.history, self.callback = _hist, _callback
            self.history.append(obj)

    # =========================================================================================================
    # TO BE OVERWRITTEN
    # =========================================================================================================

    def _setup(self, problem, **kwargs):
        pass

    def _initialize_infill(self):
        pass

    def _initialize_advance(self, infills=None, **kwargs):
        pass

    def _infill(self):
        pass

    def _advance(self, infills=None, **kwargs):
        pass

    def _finalize(self):
        pass



import json
import os
import pickle
from copy import deepcopy
from math import ceil

import numpy as np

from pymoo.experimental.benchmarking.util import at_least2d
from pymoo.optimize import minimize
from pymoo.util.misc import from_dict


class Benchmark:

    def __init__(self,
                 recorder=None,
                 n_runs=11,
                 default_termination=None):
        super().__init__()

        self.default_callback = recorder
        self.default_runs = n_runs
        self.default_termination = default_termination

        self.writer = None
        self.loader = None
        self.extractor = None

        self.problems = {}
        self.runs = []

        self.results = []

    def add_problem(self, label, problem, termination=None):
        self.problems[label] = dict(label=label, obj=problem, termination=termination)

    def add_algorithm(self, label, algorithm, problems=None, n_runs=None, termination=None, callback=None):

        if callback is None:
            callback = self.default_callback

        if problems is None:

            if len(self.problems) == 0:
                raise Exception("No Problem have been added to run the algorithm on!")

            problems = list(self.problems.keys())
        elif not isinstance(problems, list):
            problems = [problems]

        if n_runs is None:
            n_runs = self.default_runs

        for problem in problems:

            if termination is None:
                termination = self.problems[problem]["termination"]

            if termination is None:
                termination = self.default_termination

            for run in range(1, n_runs + 1):
                args = dict(problem=self.problems[problem]["obj"], algorithm=algorithm, termination=termination,
                            seed=run, callback=callback)
                e = dict(problem=problem, algorithm=label, run=run, args=args)
                self.runs.append(e)

    def run(self, ordered_by="problem", batch=None, verbose=True, **kwargs):

        params = self.runs
        params.sort(key=lambda x: x[ordered_by])

        if batch is not None:
            batch, n_batches = [int(e) for e in batch.split("/")]
            assert 0 < batch <= n_batches, f"Batch number must be greater than 0 but less or equal to number of batches!"

            batch_size = ceil(len(params) / n_batches)
            i, j = (batch - 1) * batch_size, batch * batch_size
            params = params[i:j]

        if "extractor" not in kwargs:
            kwargs["extractor"] = DefaultExtractor()

        self.results = run_looped(params, batch=batch, verbose=verbose, **kwargs)
        # self.results = run_parallel(kwargs.get("starmap"), self, params, batch=batch, verbose=verbose, **kwargs)

        return self.results


# ---------------------------------------------------------------------------------------------------------
# Functions
# ---------------------------------------------------------------------------------------------------------


def run_looped(params,
               batch,
               verbose=False,
               loader=None,
               writer=None,
               extractor=None,
               mem_free=False,
               run_if_loading_fails=True,
               exception_if_not_available=False,
               **kwargs):
    ret = []
    for i, param in enumerate(params):

        problem, algorithm, run, args = from_dict(param, "problem", "algorithm", "run", "args")

        if verbose:
            line = f"{i + 1}/{len(params)} | {algorithm} | {problem} | {run}"
            if batch is not None:
                line = f"{batch} | {line}"
            print(line, end="")

        entry = None

        if loader is not None:
            entry = loader.load(param)
            if entry is not None:
                if verbose:
                    print(f" | Loaded")

        if entry is None:
            if run_if_loading_fails:

                res = execute(param)

                entry = extractor.extract(param, res)

                if verbose:
                    print(f" | {np.round(res.exec_time, 6)} s")

                if writer is not None:
                    writer.write(entry)

            else:
                if exception_if_not_available:
                    assert entry is not None, f"Error while loading {param}"
                if verbose:
                    print(f" | Failed")

        # if the algorithm should not keep anything in memory and just write files do that
        if entry is not None and not mem_free:
            ret.append(entry)

    return ret


def execute(param):
    args = deepcopy(param['args'])

    problem = args.pop("problem")
    algorithm = args.pop("algorithm")
    res = minimize(problem, algorithm, return_least_infeasible=True, **args)
    return res


class IO:

    def __init__(self, folder) -> None:
        super().__init__()
        self.folder = folder


class DefaultWriter(IO):

    def write(self, entry):

        problem, algorithm, run, callback = from_dict(entry, "problem", "algorithm", "run", "callback")

        folder = os.path.join(self.folder, algorithm, problem)
        if not os.path.exists(folder):
            os.makedirs(folder)

        for key in ["X", "CV", "F"]:
            np.savetxt(os.path.join(folder, f"{run}.{key.lower()}"), entry.get(key))

        path = os.path.join(folder, f"{run}.json")
        with open(path, 'w') as f:
            json.dump(entry.get("info"), f, ensure_ascii=False, indent=4)

        if callback is not None:
            pickle.dump(callback, open(os.path.join(folder, f"{run}.dat"), 'wb'))


class DefaultLoader(IO):

    def load(self, entry):

        problem, algorithm, run = from_dict(entry, "problem", "algorithm", "run")

        entry = dict(entry)
        del entry["args"]

        path = os.path.join(self.folder, algorithm, problem)
        if os.path.exists(path):

            for key in ["X", "CV", "F"]:
                file = os.path.join(path, f"{run}.{key.lower()}")
                if os.path.exists(file):
                    try:
                        vals = np.loadtxt(file)
                    except:
                        return None
                    if len(vals.shape) == 0:
                        vals = np.array([float(vals)])
                    entry[key] = at_least2d(vals, expand="r")
                else:
                    return None

            file = os.path.join(path, f"{run}.dat")
            if os.path.exists(file):
                entry["callback"] = pickle.load(open(file, 'rb'))

            return entry


class DefaultExtractor:

    def extract(self, param, res):
        opt = res.opt

        entry = {
            "problem": param["problem"],
            "algorithm": param["algorithm"],
            "run": param["run"],
            "X": opt.get("X"),
            "CV": opt.get("CV"),
            "F": opt.get("F"),
            "info": dict(time=res.exec_time),
        }

        if res.algorithm.callback is not None:
            entry["callback"] = res.algorithm.callback.data,

        return entry



import re

from pymoo.config import Config
from pymoo.problems.many import *
from pymoo.problems.multi import *
from pymoo.problems.single import *


# =========================================================================================================
# Generic
# =========================================================================================================


def get_from_list(l, name, args, kwargs):
    i = None

    for k, e in enumerate(l):
        if e[0] == name:
            i = k
            break

    if i is None:
        for k, e in enumerate(l):
            if re.match(e[0], name):
                i = k
                break

    if i is not None:

        if len(l[i]) == 2:
            name, clazz = l[i]

        elif len(l[i]) == 3:
            name, clazz, default_kwargs = l[i]

            # overwrite the default if provided
            for key, val in kwargs.items():
                default_kwargs[key] = val
            kwargs = default_kwargs

        return clazz(*args, **kwargs)
    else:
        raise Exception("Object '%s' for not found in %s" % (name, [e[0] for e in l]))


# =========================================================================================================
# Algorithms
# =========================================================================================================

def get_algorithm_options():
    from pymoo.algorithms.moo.ctaea import CTAEA
    from pymoo.algorithms.moo.moead import MOEAD
    from pymoo.algorithms.moo.nsga2 import NSGA2
    from pymoo.algorithms.moo.nsga3 import NSGA3
    from pymoo.algorithms.moo.rnsga2 import RNSGA2
    from pymoo.algorithms.moo.rnsga3 import RNSGA3
    from pymoo.algorithms.soo.nonconvex.de import DE
    from pymoo.algorithms.soo.nonconvex.ga import GA
    from pymoo.algorithms.moo.unsga3 import UNSGA3
    from pymoo.algorithms.soo.nonconvex.nelder_mead import NelderMead
    from pymoo.algorithms.soo.nonconvex.cmaes import CMAES
    from pymoo.algorithms.soo.nonconvex.brkga import BRKGA
    from pymoo.algorithms.soo.nonconvex.pattern_search import PatternSearch
    from pymoo.algorithms.soo.nonconvex.pso import PSO

    ALGORITHMS = [
        ("ga", GA),
        ("brkga", BRKGA),
        ("de", DE),
        ("nelder-mead", NelderMead),
        ("pattern-search", PatternSearch),
        ("cmaes", CMAES),
        ("pso", PSO),
        ("nsga2", NSGA2),
        ("rnsga2", RNSGA2),
        ("nsga3", NSGA3),
        ("unsga3", UNSGA3),
        ("rnsga3", RNSGA3),
        ("moead", MOEAD),
        ("ctaea", CTAEA),
    ]

    return ALGORITHMS


def get_algorithm(name, *args, d={}, **kwargs):
    return get_from_list(get_algorithm_options(), name, args, {**d, **kwargs})


# =========================================================================================================
# Sampling
# =========================================================================================================

def get_sampling_options():
    from pymoo.operators.sampling.lhs import LatinHypercubeSampling
    from pymoo.operators.sampling.rnd import FloatRandomSampling
    from pymoo.operators.integer_from_float_operator import IntegerFromFloatSampling
    from pymoo.operators.sampling.rnd import BinaryRandomSampling
    from pymoo.operators.sampling.rnd import PermutationRandomSampling

    SAMPLING = [
        ("real_random", FloatRandomSampling),
        ("real_lhs", LatinHypercubeSampling),
        ("bin_random", BinaryRandomSampling),
        ("int_random", IntegerFromFloatSampling, {'clazz': FloatRandomSampling}),
        ("int_lhs", IntegerFromFloatSampling, {'clazz': LatinHypercubeSampling}),
        ("perm_random", PermutationRandomSampling)
    ]

    return SAMPLING


def get_sampling(name, *args, d={}, **kwargs):
    return get_from_list(get_sampling_options(), name, args, {**d, **kwargs})


# =========================================================================================================
# Selection
# =========================================================================================================

def get_selection_options():
    from pymoo.operators.selection.rnd import RandomSelection
    from pymoo.operators.selection.tournament import TournamentSelection

    SELECTION = [
        ("random", RandomSelection),
        ("tournament", TournamentSelection)
    ]

    return SELECTION


def get_selection(name, *args, d={}, **kwargs):
    return get_from_list(get_selection_options(), name, args, {**d, **kwargs})


# =========================================================================================================
# Crossover
# =========================================================================================================

def get_crossover_options():
    from pymoo.operators.crossover.dex import DEX
    from pymoo.operators.crossover.expx import ExponentialCrossover
    from pymoo.operators.crossover.hux import HalfUniformCrossover
    from pymoo.operators.crossover.pntx import PointCrossover
    from pymoo.operators.crossover.sbx import SimulatedBinaryCrossover
    from pymoo.operators.crossover.ux import UniformCrossover
    from pymoo.operators.crossover.pcx import PCX
    from pymoo.operators.integer_from_float_operator import IntegerFromFloatCrossover
    from pymoo.operators.crossover.erx import EdgeRecombinationCrossover
    from pymoo.operators.crossover.ox import OrderCrossover

    CROSSOVER = [
        ("real_sbx", SimulatedBinaryCrossover, dict(prob=0.9, eta=30)),
        ("int_sbx", IntegerFromFloatCrossover, dict(clazz=SimulatedBinaryCrossover, prob=0.9, eta=30)),
        ("real_de", DEX),
        ("real_pcx", PCX),
        ("(real|bin|int)_ux", UniformCrossover),
        ("(bin|int)_hux", HalfUniformCrossover),
        ("(real|bin|int)_exp", ExponentialCrossover),
        ("(real|bin|int)_one_point", PointCrossover, {'n_points': 1}),
        ("(real|bin|int)_two_point", PointCrossover, {'n_points': 2}),
        ("(real|bin|int)_k_point", PointCrossover),
        ("perm_ox", OrderCrossover),
        ("perm_erx", EdgeRecombinationCrossover)
    ]

    return CROSSOVER


def get_crossover(name, *args, d={}, **kwargs):
    return get_from_list(get_crossover_options(), name, args, {**d, **kwargs})


# =========================================================================================================
# Mutation
# =========================================================================================================

def get_mutation_options():
    from pymoo.operators.mutation.nom import NoMutation
    from pymoo.operators.mutation.bitflip import BinaryBitflipMutation
    from pymoo.operators.mutation.pm import PolynomialMutation
    from pymoo.operators.integer_from_float_operator import IntegerFromFloatMutation
    from pymoo.operators.mutation.inversion import InversionMutation

    MUTATION = [
        ("none", NoMutation, {}),
        ("real_pm", PolynomialMutation, dict(eta=20)),
        ("int_pm", IntegerFromFloatMutation, dict(clazz=PolynomialMutation, eta=20)),
        ("bin_bitflip", BinaryBitflipMutation),
        ("perm_inv", InversionMutation)
    ]

    return MUTATION


def get_mutation(name, *args, d={}, **kwargs):
    return get_from_list(get_mutation_options(), name, args, {**d, **kwargs})


# =========================================================================================================
# Termination
# =========================================================================================================

def get_termination_options():
    from pymoo.util.termination.max_eval import MaximumFunctionCallTermination
    from pymoo.util.termination.max_gen import MaximumGenerationTermination
    from pymoo.util.termination.max_time import TimeBasedTermination
    from pymoo.util.termination.min_igd import IGDTermination
    from pymoo.util.termination.x_tol import DesignSpaceToleranceTermination
    from pymoo.util.termination.f_tol import MultiObjectiveSpaceToleranceTermination
    from pymoo.util.termination.f_tol_single import SingleObjectiveSpaceToleranceTermination
    from pymoo.util.termination.default import MultiObjectiveDefaultTermination, SingleObjectiveDefaultTermination

    TERMINATION = [
        ("n_eval", MaximumFunctionCallTermination),
        ("(n_gen|n_iter)", MaximumGenerationTermination),
        ("igd", IGDTermination),
        ("time", TimeBasedTermination),
        ("(x_tol|xtol)", DesignSpaceToleranceTermination),
        ("(f_tol$|ftol$)", MultiObjectiveSpaceToleranceTermination),
        ("(f_tol_s|ftol_s)", SingleObjectiveSpaceToleranceTermination),
        ("(default$|default_multi)", MultiObjectiveDefaultTermination),
        ("default_single$", SingleObjectiveDefaultTermination)
    ]

    return TERMINATION


def get_termination(name, *args, d={}, **kwargs):
    return get_from_list(get_termination_options(), name, args, {**d, **kwargs})


# =========================================================================================================
# Problems
# =========================================================================================================

def get_problem_options():
    PROBLEM = [
        ('ackley', Ackley),
        ('bnh', BNH),
        ('carside', Carside),
        ('ctp1', CTP1),
        ('ctp2', CTP2),
        ('ctp3', CTP3),
        ('ctp4', CTP4),
        ('ctp5', CTP5),
        ('ctp6', CTP6),
        ('ctp7', CTP7),
        ('ctp8', CTP8),
        ('dascmop1', DASCMOP1),
        ('dascmop2', DASCMOP2),
        ('dascmop3', DASCMOP3),
        ('dascmop4', DASCMOP4),
        ('dascmop5', DASCMOP5),
        ('dascmop6', DASCMOP6),
        ('dascmop7', DASCMOP7),
        ('dascmop8', DASCMOP8),
        ('dascmop9', DASCMOP9),
        ('modact', MODAct),
        ('mw1', MW1),
        ('mw2', MW2),
        ('mw3', MW3),
        ('mw4', MW4),
        ('mw5', MW5),
        ('mw6', MW6),
        ('mw7', MW7),
        ('mw8', MW8),
        ('mw9', MW9),
        ('mw10', MW10),
        ('mw11', MW11),
        ('mw12', MW12),
        ('mw13', MW13),
        ('mw14', MW14),
        ('dtlz1^-1', InvertedDTLZ1),
        ('dtlz1', DTLZ1),
        ('dtlz2', DTLZ2),
        ('dtlz3', DTLZ3),
        ('dtlz4', DTLZ4),
        ('dtlz5', DTLZ5),
        ('dtlz6', DTLZ6),
        ('dtlz7', DTLZ7),
        ('convex_dtlz2', ConvexDTLZ2),
        ('convex_dtlz4', ConvexDTLZ4),
        ('sdtlz1', ScaledDTLZ1),
        ('c1dtlz1', C1DTLZ1),
        ('c1dtlz3', C1DTLZ3),
        ('c2dtlz2', C2DTLZ2),
        ('c3dtlz1', C3DTLZ1),
        ('c3dtlz4', C3DTLZ4),
        ('dc1dtlz1', DC1DTLZ1),
        ('dc1dtlz3', DC1DTLZ3),
        ('dc2dtlz1', DC2DTLZ1),
        ('dc2dtlz3', DC2DTLZ3),
        ('dc3dtlz1', DC3DTLZ1),
        ('dc3dtlz3', DC3DTLZ3),
        ('cantilevered_beam', CantileveredBeam),
        ('griewank', Griewank),
        ('himmelblau', Himmelblau),
        ('knp', Knapsack),
        ('kursawe', Kursawe),
        ('osy', OSY),
        ('pressure_vessel', PressureVessel),
        ('rastrigin', Rastrigin),
        ('rosenbrock', Rosenbrock),
        ('schwefel', Schwefel),
        ('sphere', Sphere),
        ('srn', SRN),
        ('tnk', TNK),
        ('truss2d', Truss2D),
        ('welded_beam', WeldedBeam),
        ('zakharov', Zakharov),
        ('zdt1', ZDT1),
        ('zdt2', ZDT2),
        ('zdt3', ZDT3),
        ('zdt4', ZDT4),
        ('zdt5', ZDT5),
        ('zdt6', ZDT6),
        ('g01', G1),
        ('g02', G2),
        ('g03', G3),
        ('g04', G4),
        ('g05', G5),
        ('g06', G6),
        ('g07', G7),
        ('g08', G8),
        ('g09', G9),
        ('g10', G10),
        ('wfg1', WFG1),
        ('wfg2', WFG2),
        ('wfg3', WFG3),
        ('wfg4', WFG4),
        ('wfg5', WFG5),
        ('wfg6', WFG6),
        ('wfg7', WFG7),
        ('wfg8', WFG8),
        ('wfg9', WFG9)
    ]

    return PROBLEM


def get_problem(name, *args, d={}, **kwargs):
    if name.startswith("go-"):
        from pymoo.vendor.global_opt import get_global_optimization_problem_options
        return get_from_list(get_global_optimization_problem_options(), name.lower(), args, {**d, **kwargs})
    elif name.startswith("bbob-"):
        from pymoo.vendor.vendor_coco import COCOProblem
        return COCOProblem(name.lower(), **kwargs)
    else:
        return get_from_list(get_problem_options(), name.lower(), args, {**d, **kwargs})


# =========================================================================================================
# Weights
# =========================================================================================================

def get_reference_direction_options():
    from pymoo.util.reference_direction import UniformReferenceDirectionFactory
    from pymoo.util.reference_direction import MultiLayerReferenceDirectionFactory
    from pymoo.util.ref_dirs.reduction import ReductionBasedReferenceDirectionFactory
    from pymoo.util.ref_dirs.energy import RieszEnergyReferenceDirectionFactory
    from pymoo.util.ref_dirs.energy_layer import LayerwiseRieszEnergyReferenceDirectionFactory

    REFERENCE_DIRECTIONS = [
        ("(das-dennis|uniform)", UniformReferenceDirectionFactory),
        ("multi-layer", MultiLayerReferenceDirectionFactory),
        ("(energy|riesz)", RieszEnergyReferenceDirectionFactory),
        ("(layer-energy|layer-riesz)", LayerwiseRieszEnergyReferenceDirectionFactory),
        ("red", ReductionBasedReferenceDirectionFactory)
    ]

    return REFERENCE_DIRECTIONS


def get_reference_directions(name, *args, d={}, **kwargs):
    return get_from_list(get_reference_direction_options(), name, args, {**d, **kwargs}).do()


# =========================================================================================================
# Visualization
# =========================================================================================================

def get_visualization_options():
    from pymoo.visualization.pcp import PCP
    from pymoo.visualization.petal import Petal
    from pymoo.visualization.radar import Radar
    from pymoo.visualization.radviz import Radviz
    from pymoo.visualization.scatter import Scatter
    from pymoo.visualization.star_coordinate import StarCoordinate
    from pymoo.visualization.heatmap import Heatmap
    from pymoo.visualization.fitness_landscape import FitnessLandscape

    VISUALIZATION = [
        ("scatter", Scatter),
        ("heatmap", Heatmap),
        ("pcp", PCP),
        ("petal", Petal),
        ("radar", Radar),
        ("radviz", Radviz),
        ("star", StarCoordinate),
        ("fitness-landscape", FitnessLandscape)
    ]

    return VISUALIZATION


def get_visualization(name, *args, d={}, **kwargs):
    return get_from_list(get_visualization_options(), name, args, {**d, **kwargs})


# =========================================================================================================
# Performance Indicator
# =========================================================================================================


def get_performance_indicator_options():
    from pymoo.indicators.gd import GD
    from pymoo.indicators.gd_plus import GDPlus
    from pymoo.indicators.igd import IGD
    from pymoo.indicators.igd_plus import IGDPlus
    from pymoo.indicators.hv import Hypervolume
    from pymoo.indicators.rmetric import RMetric

    PERFORMANCE_INDICATOR = [
        ("gd", GD),
        ("gd+", GDPlus),
        ("igd", IGD),
        ("igd+", IGDPlus),
        ("hv", Hypervolume),
        ("rmetric", RMetric)
    ]
    return PERFORMANCE_INDICATOR


def get_performance_indicator(name, *args, d={}, **kwargs):
    return get_from_list(get_performance_indicator_options(), name, args, {**d, **kwargs})


# =========================================================================================================
# DECOMPOSITION
# =========================================================================================================

def get_decomposition_options():
    from pymoo.decomposition.pbi import PBI
    from pymoo.decomposition.tchebicheff import Tchebicheff
    from pymoo.decomposition.weighted_sum import WeightedSum
    from pymoo.decomposition.asf import ASF
    from pymoo.decomposition.aasf import AASF
    from pymoo.decomposition.perp_dist import PerpendicularDistance

    DECOMPOSITION = [
        ("weighted-sum", WeightedSum),
        ("tchebi", Tchebicheff),
        ("pbi", PBI),
        ("asf", ASF),
        ("aasf", AASF),
        ("perp_dist", PerpendicularDistance)
    ]

    return DECOMPOSITION


def get_decomposition(name, *args, d={}, **kwargs):
    return get_from_list(get_decomposition_options(), name, args, {**d, **kwargs})


# =========================================================================================================
# DECOMPOSITION
# =========================================================================================================

def get_decision_making_options():
    from pymoo.mcdm.high_tradeoff import HighTradeoffPoints
    from pymoo.mcdm.pseudo_weights import PseudoWeights

    DECISION_MAKING = [
        ("high-tradeoff", HighTradeoffPoints),
        ("pseudo-weights", PseudoWeights)
    ]

    return DECISION_MAKING


def get_decision_making(name, *args, d={}, **kwargs):
    return get_from_list(get_decision_making_options(), name, args, {**d, **kwargs})


# =========================================================================================================
# Documentation
# =========================================================================================================


def dummy(name, kwargs):
    """
    A convenience method to get a {type} object just by providing a string.

    Parameters
    ----------

    name : {{ {options} }}
        Name of the {type}.

    kwargs : dict
        Dictionary that should be used to call the method mapped to the {type} factory function.

    Returns
    -------
    class : {clazz}
        An {type} object based on the string. `None` if the {type} was not found.

    """
    pass


def options_to_string(l):
    return ", ".join(["'%s'" % k[0] for k in l])


if Config.parse_custom_docs:
    from pymoo.docs import parse_doc_string

    from pymoo.factory import get_algorithm_options, get_selection_options, get_crossover_options, \
        get_mutation_options, get_termination_options, get_algorithm, get_selection, get_crossover, get_mutation, \
        get_termination, get_sampling, get_sampling_options

    parse_doc_string(dummy, get_algorithm, {"type": "algorithm",
                                            "clazz": ":class:`~pymoo.core.algorithm.Algorithm`",
                                            "options": options_to_string(get_algorithm_options())
                                            })

    parse_doc_string(dummy, get_sampling, {"type": "sampling",
                                           "clazz": ":class:`~pymoo.core.sampling.Sampling`",
                                           "options": options_to_string(get_sampling_options())
                                           })

    parse_doc_string(dummy, get_selection, {"type": "selection",
                                            "clazz": ":class:`~pymoo.core.selection.Selection`",
                                            "options": options_to_string(get_selection_options())
                                            })

    parse_doc_string(dummy, get_crossover, {"type": "crossover",
                                            "clazz": ":class:`~pymoo.core.crossover.Crossover`",
                                            "options": options_to_string(get_crossover_options())
                                            })

    parse_doc_string(dummy, get_mutation, {"type": "mutation",
                                           "clazz": ":class:`~pymoo.core.mutation.Mutation`",
                                           "options": options_to_string(get_mutation_options())
                                           })

    parse_doc_string(dummy, get_termination, {"type": "termination",
                                              "clazz": ":class:`~pymoo.core.termination.termination`",
                                              "options": options_to_string(get_termination_options())
                                              })




import math

import numpy as np
from scipy.spatial.distance import cdist, pdist, squareform

from pymoo.algorithms.base.genetic import GeneticAlgorithm
from pymoo.decomposition.asf import ASF
from pymoo.docs import parse_doc_string
from pymoo.core.population import Population
from pymoo.operators.crossover.sbx import SimulatedBinaryCrossover
from pymoo.operators.mutation.pm import PolynomialMutation
from pymoo.operators.sampling.rnd import FloatRandomSampling
from pymoo.operators.selection.tournament import TournamentSelection
from pymoo.util.display import MultiObjectiveDisplay
from pymoo.util.dominator import Dominator
from pymoo.util.function_loader import load_function
from pymoo.util.misc import has_feasible, random_permuations
from pymoo.util.nds.non_dominated_sorting import NonDominatedSorting


# =========================================================================================================
# Implementation
# Following original code by K. Li https://cola-laboratory.github.io/codes/CTAEA.zip
# =========================================================================================================


def comp_by_cv_dom_then_random(pop, P, **kwargs):
    S = np.full(P.shape[0], np.nan)

    for i in range(P.shape[0]):
        a, b = P[i, 0], P[i, 1]

        if pop[a].CV <= 0.0 and pop[b].CV <= 0.0:
            rel = Dominator.get_relation(pop[a].F, pop[b].F)
            if rel == 1:
                S[i] = a
            elif rel == -1:
                S[i] = b
            else:
                S[i] = np.random.choice([a, b])
        elif pop[a].CV <= 0.0:
            S[i] = a
        elif pop[b].CV <= 0.0:
            S[i] = b
        else:
            S[i] = np.random.choice([a, b])

    return S[:, None].astype(int)


class RestrictedMating(TournamentSelection):
    """Restricted mating approach to balance convergence and diversity archives"""

    def _do(self, Hm, n_select, n_parents=2, **kwargs):
        n_pop = len(Hm) // 2

        _, rank = NonDominatedSorting().do(Hm.get('F'), return_rank=True)

        Pc = (rank[:n_pop] == 0).sum() / len(Hm)
        Pd = (rank[n_pop:] == 0).sum() / len(Hm)

        # number of random individuals needed
        n_random = n_select * n_parents * self.pressure
        n_perms = math.ceil(n_random / n_pop)
        # get random permutations and reshape them
        P = random_permuations(n_perms, n_pop)[:n_random]
        P = np.reshape(P, (n_select * n_parents, self.pressure))
        if Pc <= Pd:
            # Choose from DA
            P[::n_parents, :] += n_pop
        pf = np.random.random(n_select)
        P[1::n_parents, :][pf >= Pc] += n_pop

        # compare using tournament function
        S = self.f_comp(Hm, P, **kwargs)

        return np.reshape(S, (n_select, n_parents))


class CADASurvival:

    def __init__(self, ref_dirs):
        self.ref_dirs = ref_dirs
        self.opt = None
        self.ideal_point = np.full(ref_dirs.shape[1], np.inf)
        self._decomposition = ASF()
        self._calc_perpendicular_distance = load_function("calc_perpendicular_distance")

    def do(self, _, pop, da, n_survive=None, **kwargs):
        # Offspring are last of merged population
        off = pop[-n_survive:]
        # Update ideal point
        self.ideal_point = np.min(np.vstack((self.ideal_point, off.get("F"))), axis=0)
        # Update CA
        pop = self._updateCA(pop, n_survive)
        # Update DA
        Hd = Population.merge(da, off)
        da = self._updateDA(pop, Hd, n_survive)
        return pop, da

    def _associate(self, pop):
        """Associate each individual with a F vector and calculate decomposed fitness"""
        F = pop.get("F")
        dist_matrix = self._calc_perpendicular_distance(F - self.ideal_point, self.ref_dirs)
        niche_of_individuals = np.argmin(dist_matrix, axis=1)
        FV = self._decomposition.do(F, weights=self.ref_dirs[niche_of_individuals, :],
                                    ideal_point=self.ideal_point, weight_0=1e-4)
        pop.set("niche", niche_of_individuals)
        pop.set("FV", FV)
        return niche_of_individuals, FV

    def _updateCA(self, pop, n_survive):
        """Update the Convergence archive (CA)"""
        CV = pop.get("CV").flatten()

        Sc = pop[CV == 0]  # Feasible population
        if len(Sc) == n_survive:  # Exactly n_survive feasible individuals
            F = Sc.get("F")
            fronts, rank = NonDominatedSorting().do(F, return_rank=True)
            Sc.set('rank', rank)
            self.opt = Sc[fronts[0]]
            return Sc
        elif len(Sc) < n_survive:  # Not enough feasible individuals
            remainder = n_survive - len(Sc)
            # Solve sub-problem CV, tche
            SI = pop[CV > 0]
            f1 = SI.get("CV")
            _, f2 = self._associate(SI)
            sub_F = np.column_stack([f1, f2])
            fronts = NonDominatedSorting().do(sub_F, n_stop_if_ranked=remainder)
            I = []
            for front in fronts:
                if len(I) + len(front) <= remainder:
                    I.extend(front)
                else:
                    n_missing = remainder - len(I)
                    last_front_CV = np.argsort(f1.flatten()[front])
                    I.extend(front[last_front_CV[:n_missing]])
            SI = SI[I]
            S = Population.merge(Sc, SI)
            F = S.get("F")
            fronts, rank = NonDominatedSorting().do(F, return_rank=True)
            S.set('rank', rank)
            self.opt = S[fronts[0]]
            return S
        else:  # Too many feasible individuals
            F = Sc.get("F")
            # Filter by non-dominated sorting
            fronts, rank = NonDominatedSorting().do(F, return_rank=True, n_stop_if_ranked=n_survive)
            I = np.concatenate(fronts)
            S, rank, F = Sc[I], rank[I], F[I]
            if len(S) > n_survive:
                # Remove individual in most crowded niche and with worst fitness
                niche_of_individuals, FV = self._associate(S)
                index, count = np.unique(niche_of_individuals, return_counts=True)
                survivors = np.full(S.shape[0], True)
                while survivors.sum() > n_survive:
                    crowdest_niches, = np.where(count == count.max())
                    worst_idx = None
                    worst_niche = None
                    worst_fit = -1
                    for crowdest_niche in crowdest_niches:
                        crowdest, = np.where((niche_of_individuals == index[crowdest_niche]) & survivors)
                        niche_worst = crowdest[FV[crowdest].argmax()]
                        dist_to_max_fit = cdist(F[[niche_worst], :], F).flatten()
                        dist_to_max_fit[niche_worst] = np.inf
                        dist_to_max_fit[~survivors] = np.inf
                        min_d_to_max_fit = dist_to_max_fit.min()

                        dist_in_niche = squareform(pdist(F[crowdest]))
                        np.fill_diagonal(dist_in_niche, np.inf)

                        delta_d = dist_in_niche - min_d_to_max_fit
                        min_d_i = np.unravel_index(np.argmin(delta_d, axis=None), dist_in_niche.shape)
                        if (delta_d[min_d_i] < 0) or (
                                delta_d[min_d_i] == 0 and (FV[crowdest[list(min_d_i)]] > niche_worst).any()):
                            min_d_i = list(min_d_i)
                            np.random.shuffle(min_d_i)
                            closest = crowdest[min_d_i]
                            niche_worst = closest[np.argmax(FV[closest])]
                        if FV[niche_worst] > worst_fit:
                            worst_fit = FV[niche_worst]
                            worst_idx = niche_worst
                            worst_niche = crowdest_niche
                    survivors[worst_idx] = False
                    count[worst_niche] -= 1
                S, rank = S[survivors], rank[survivors]
            S.set('rank', rank)
            self.opt = S[rank == 0]
            return S

    def _updateDA(self, pop, Hd, n_survive):
        """Update the Diversity archive (DA)"""
        niche_Hd, FV = self._associate(Hd)
        niche_CA, _ = self._associate(pop)

        itr = 1
        S = []
        while len(S) < n_survive:
            for i in range(n_survive):
                current_ca, = np.where(niche_CA == i)
                if len(current_ca) < itr:
                    for _ in range(itr - len(current_ca)):
                        current_da = np.where(niche_Hd == i)[0]
                        if current_da.size > 0:
                            F = Hd[current_da].get('F')
                            nd = NonDominatedSorting().do(F, only_non_dominated_front=True, n_stop_if_ranked=0)
                            i_best = current_da[nd[np.argmin(FV[current_da[nd]])]]
                            niche_Hd[i_best] = -1
                            if len(S) < n_survive:
                                S.append(i_best)
                        else:
                            break
                if len(S) == n_survive:
                    break
            itr += 1
        return Hd[S]




class CTAEA(GeneticAlgorithm):

    def __init__(self,
                 ref_dirs,
                 sampling=FloatRandomSampling(),
                 selection=RestrictedMating(func_comp=comp_by_cv_dom_then_random),
                 crossover=SimulatedBinaryCrossover(n_offsprings=1, eta=30, prob=1.0),
                 mutation=PolynomialMutation(eta=20, prob=None),
                 eliminate_duplicates=True,
                 display=MultiObjectiveDisplay(),
                 **kwargs):
        """
        CTAEA

        Parameters
        ----------
        ref_dirs : {ref_dirs}
        sampling : {sampling}
        selection : {selection}
        crossover : {crossover}
        mutation : {mutation}
        eliminate_duplicates : {eliminate_duplicates}
        """

        self.ref_dirs = ref_dirs
        pop_size = len(ref_dirs)

        if 'survival' in kwargs:
            survival = kwargs['survival']
            del kwargs['survival']
        else:
            survival = CADASurvival(ref_dirs)

        # Initialize diversity archives
        self.da = None

        super().__init__(pop_size=pop_size,
                         sampling=sampling,
                         selection=selection,
                         crossover=crossover,
                         mutation=mutation,
                         survival=survival,
                         eliminate_duplicates=eliminate_duplicates,
                         n_offsprings=pop_size,
                         display=display,
                         **kwargs)

    def _setup(self, problem, **kwargs):

        if self.ref_dirs is not None and self.ref_dirs.shape[1] != problem.n_obj:
            raise Exception(
                "Dimensionality of reference points must be equal to the number of objectives: %s != %s" %
                (self.ref_dirs.shape[1], problem.n_obj))

    def _initialize_infill(self):
        return self.initialization.do(self.problem, self.pop_size, algorithm=self)

    def _initialize_advance(self, infills=None, **kwargs):
        super()._initialize_advance(infills, **kwargs)
        self.pop, self.da = self.survival.do(self.problem, self.pop, Population(), n_survive=len(self.pop),
                                             algorithm=self)

    def _infill(self):
        Hm = Population.merge(self.pop, self.da)
        return self.mating.do(self.problem, Hm, n_offsprings=self.n_offsprings, algorithm=self)

    def _advance(self, infills=None, **kwargs):
        assert infills is not None, "This algorithms uses the AskAndTell interface thus infills must to be provided."
        pop = Population.merge(self.pop, infills)
        self.pop, self.da = self.survival.do(self.problem, pop, self.da, self.pop_size, algorithm=self)

    def _set_optimum(self, **kwargs):
        if not has_feasible(self.pop):
            self.opt = self.pop[[np.argmin(self.pop.get("CV"))]]
        else:
            self.opt = self.survival.opt


parse_doc_string(CTAEA.__init__)



import numpy as np
from scipy.linalg import solve_triangular, LinAlgError

from pymoo.algorithms.base.gradient import DerivationBasedAlgorithm, inexact_line_search
from pymoo.algorithms.base.line import LineSearchProblem
from pymoo.algorithms.soo.univariate.wolfe import wolfe_line_search, WolfeSearch
from pymoo.core.individual import Individual
from pymoo.core.population import Population
from pymoo.core.solution import Solution


def direction_cholesky(jac, hess):
    # https://web.stanford.edu/~boyd/cvxbook/bv_cvxbook.pdf, Page: 510
    adapted = False

    try:
        L = np.linalg.cholesky(hess)

    # hessian is not positive definite, we fix it (for degenerated cases this may lead to "bad" directions)
    except LinAlgError:
        eigvals, _ = np.linalg.eig(hess)
        E = max(0, 1e-5 - eigvals.min()) * np.eye(len(jac))
        try:
            L = np.linalg.cholesky(hess + E)
        # this should happen for very degnerated cases only
        except LinAlgError:
            return None, None, None, False
        adapted = True

    w = solve_triangular(L, - jac, lower=True)
    dir = solve_triangular(L.T, w, lower=False)
    dec = np.linalg.norm(w) ** 2
    return dir, dec, adapted, True


def direction_inv(jac, hess):
    try:
        hess_inv = np.linalg.inv(hess)
        dir = - hess_inv @ jac
        dec = jac.T @ hess_inv @ jac
    except:
        return None, None, False, False
    return dir, dec, False, True




class NewtonMethod(DerivationBasedAlgorithm):

    def __init__(self, damped=True, eps=1e-8, **kwargs):
        super().__init__(**kwargs)
        self.damped = damped
        self.eps = eps

    def setup(self, problem, **kwargs):
        super().setup(problem, **kwargs)
        assert problem.n_obj == 1, "The Newton method can only be used for single-objective problems (n_obj=1)."
        assert problem.n_constr == 0, "The Newton method can only be used for unconstrained problems (n_constr=0)."
        return self

    def _next(self):
        sol = self.opt[0]
        x = sol.get("X")
        jac, hess = self.gradient_and_hessian(sol)

        method = "cholesky"
        func = direction_cholesky if method == "cholesky" else direction_inv
        direction, decrement, adapted, success = func(jac, hess)

        # in case we can not calculate the newton direction fall back to the gradient approach
        if not success:
            direction = -jac
            decrement = np.linalg.norm(direction) ** 2

        if self.damped:
            line = LineSearchProblem(self.problem, sol, direction)
            _next = WolfeSearch().setup(line, evaluator=self.evaluator).run()
            # _next = wolfe_line_search(self.problem, sol, direction, evaluator=self.evaluator)

            # if the newton step was not successful, then try the gradient
            if adapted and (sol.F[0] - _next.F[0]) < 1e-16:
                _next = inexact_line_search(self.problem, sol, -jac, evaluator=self.evaluator)

        else:
            _x = x + direction
            _next = Solution(X=_x)

        if isinstance(_next, Individual):
            _next = Population.create(_next)

        self.pop = Population.merge(self.opt, _next)

        if decrement / 2 <= self.eps:
            self.termination.force_termination = True



from pymoo.algorithms.base.line import LineSearchProblem
from pymoo.algorithms.base.local import LocalSearch
from pymoo.algorithms.soo.univariate.backtracking import BacktrackingLineSearch
from pymoo.algorithms.soo.univariate.golden import GoldenSectionSearch
from pymoo.optimize import minimize

from pymoo.util.display import SingleObjectiveDisplay
from pymoo.util.termination.f_tol_single import SingleObjectiveSpaceToleranceTermination


class DerivationBasedDisplay(SingleObjectiveDisplay):

    def _do(self, problem, evaluator, algorithm):
        super()._do(problem, evaluator, algorithm)
        self.output.append("alpha", "-" if algorithm.alpha is None else algorithm.alpha)


class DerivationBasedAlgorithm(LocalSearch):

    def __init__(self, strict_bounds=True, requires_hessian=False, display=DerivationBasedDisplay(), **kwargs):
        super().__init__(display=display, **kwargs)

        # whether the bounds are treated to be strict
        self.strict_bounds = strict_bounds

        # whether the implementation requires the calculation of the hessian matrix
        self.requires_hessian = requires_hessian

        # gradient of the solution in the current iteration
        self.dF = None

        # the direction that has been used for the line search
        self.dir = None

        # the resulting scalar from the line search
        self.alpha = 1.0

        # set the default termination
        self.default_termination = SingleObjectiveSpaceToleranceTermination(n_last=3, tol=1e-10)

    def _setup(self, problem, **kwargs):
        assert problem.n_obj == 1, "The Newton method can only be used for single-objective problems (n_obj=1)."
        assert problem.n_constr == 0, "The Newton method can only be used for unconstrained problems (n_constr=0)."



def exact_line_search(problem, sol, direction, algorithm=GoldenSectionSearch(), n_iter=10, evaluator=None):
    problem = LineSearchProblem(problem, sol, direction)
    res = minimize(problem, algorithm, ("n_iter", n_iter), evaluator=evaluator)
    sol = res.opt[0]
    sol.set("X", sol.get("__X__"))
    return sol


def inexact_line_search(problem, sol, direction, evaluator=None):
    algorithm = BacktrackingLineSearch()
    algorithm = algorithm.setup(problem, point=sol, direction=direction)._initialize()
    res = minimize(problem, algorithm, evaluator=evaluator)
    return res.opt[0]



class Cache:

    def __init__(self, func, raise_exception=True) -> None:
        super().__init__()

        # the function to be executed
        self.func = func

        # the result object which is cached
        self.cache = None

        # a flag indicating if it has been executed
        self.has_been_executed = False

        # whether an exception shall be thrown if it fails
        self.raise_exception = raise_exception

        # the error message if an exception occurs
        self.error = None

    def exec(self, *args, use_cache=True, **kwargs):

        if use_cache and self.has_been_executed:
            return self.cache

        else:
            try:
                self.cache = self.func(*args, **kwargs)
            except Exception as e:
                self.error = str(e)

                if self.raise_exception:
                    raise e

            self.has_been_executed = True

            return self.cache

    def reset(self):
        self.has_been_executed = False
        self.cache = None
        self.error = None



import numpy as np

from pymoo.core.crossover import Crossover


class BlendCrossover(Crossover):
    def __init__(self, alpha, prob_per_variable=0.5, **kwargs):
        super().__init__(2, 2, **kwargs)
        self.alpha = float(alpha)
        self.prob_per_variable = prob_per_variable

    def _do(self, problem, X, **kwargs):

        X = X.astype(float)
        _, n_matings, n_var = X.shape

        # boundaries of the problem
        xl, xu = problem.xl, problem.xu

        # crossover mask that will be used in the end
        do_crossover = np.full(X[0].shape, True)

        # per variable the probability is then 50%
        do_crossover[np.random.random((n_matings, problem.n_var)) > self.prob_per_variable] = False
        # also if values are too close no mating is done
        do_crossover[np.abs(X[0] - X[1]) <= 1.0e-14] = False

        d = np.abs(X[0, :, :] - X[1, :, :])
        y1 = np.min(X, axis=0)
        y2 = np.max(X, axis=0)

        _min = y1 - self.alpha * d
        _max = y2 + self.alpha * d

        c = np.random.uniform(_min, _max, X.shape)
        # for i in range(X.shape[1]):
        #     for j in range(X.shape[2]):
        #         c[:, i, j] = np.random.uniform(y1[i, j] - self.alpha*d[i, j], y2[i, j] + self.alpha*d[i, j], 2)

        return c



import numpy as np

from pymoo.algorithms.soo.nonconvex.de import DifferentialEvolutionMating
from pymoo.algorithms.soo.nonconvex.ga import GA
from pymoo.factory import get_problem
from pymoo.core.population import Population
from pymoo.core.replacement import hierarchical_sort
from pymoo.operators.crossover.pcx import PCX
from pymoo.operators.selection.tournament import TournamentSelection, compare
from pymoo.optimize import minimize


# =========================================================================================================
# Implementation
# =========================================================================================================


def constr_binary_tournament(pop, P, algorithm, **kwargs):
    n_tournaments, n_parents = P.shape

    if n_parents != 2:
        raise ValueError("Only implemented for binary tournament!")

    S = np.full(n_tournaments, np.nan)

    for i in range(n_tournaments):

        a, b = P[i, 0], P[i, 1]
        a_cv, a_f, b_cv, b_f, = pop[a].CV[0], pop[a].F, pop[b].CV[0], pop[b].F
        a_below, b_below = pop[a].get("below"), pop[b].get("below")

        if a_cv > 0.0 and b_cv > 0.0:
            if a_below and b_below:
                S[i] = compare(a, a_cv, b, b_cv, method='smaller_is_better', return_random_if_equal=True)
            elif not a_below and not b_below:
                S[i] = compare(a, a_f, b, b_f, method='smaller_is_better', return_random_if_equal=True)
            elif a_below and not b_below:
                S[i] = a
            elif not a_below and b_below:
                S[i] = b

        elif a_cv == 0.0 and b_cv == 0.0:
            S[i] = np.random.choice([a, b])
            # S[i] = compare(a, a_f, b, b_f, method='smaller_is_better', return_random_if_equal=True)
        else:
            S[i] = np.random.choice([a, b])

    return S[:, None].astype(int, copy=False)


class ConstrGA(GA):

    def __init__(self,
                 selection=TournamentSelection(func_comp=constr_binary_tournament),
                 **kwargs):
        super().__init__(selection=selection, advance_after_initial_infill=True, **kwargs)

    def _advance(self, infills=None, **kwargs):
        assert infills is not None, "This algorithms uses the AskAndTell interface thus infills must to be provided."

        pop = Population.merge(self.pop, infills)
        F, CV = pop.get("F", "CV")
        f, cv = F[:, 0, ], CV[:, 0]

        S = hierarchical_sort(f, cv)
        pop, f, cv, feas = pop[S], f[S], cv[S], cv[S] <= 0

        survivors = list(range(30))
        n_remaining = self.pop_size - len(survivors)

        if feas.sum() > 0:
            f_min, cv_min = f[feas].min(), 0.0
        else:
            f_min, cv_min = np.inf, cv[~feas].min()

        I = np.where(np.logical_and(~feas, f < f_min))[0]
        # survivors.extend(I[:n_remaining])
        survivors.extend(I[cv[I].argsort()][:n_remaining])

        if len(survivors) < n_remaining:
            I = np.where(np.logical_and(~feas, f >= f_min))[0]
            survivors.extend(I[f[I].argsort()][:n_remaining])
            # survivors.extend(np.random.choice(I, size=n_remaining))

        if self.n_gen > 1000:
            import matplotlib.pyplot as plt

            plt.scatter(cv[~feas], f[~feas], facecolor="none", edgecolors="red", alpha=0.5, s=20)
            plt.scatter(cv[feas], f[feas], facecolor="none", edgecolors="blue", alpha=0.5, s=40)

            plt.scatter(cv[survivors], f[survivors], color="black", s=3, alpha=0.9)

            plt.show()

        self.pop = pop[survivors]
        self.pop.set("below", self.pop.get("F")[:, 0] <= f_min)

        # print("n_feas", np.sum(pop[survivors].get("CV")[:, 0] <= 0))




problem = get_problem("g09")

algorithm = ConstrGA(pop_size=200,
                     # mating=DifferentialEvolutionMating(variant="DE/best/1/bin")
                     )

res = minimize(problem,
               algorithm,
               ("n_gen", 3000),
               seed=1,
               verbose=True)

print("Best solution found: \nX = %s\nF = %s" % (res.X, res.F))



import warnings

import autograd
import autograd.numpy as anp
from autograd.core import VJPNode, vspace, backward_pass
from autograd.tracer import new_box, isbox

from pymoo.core.problem import ElementwiseProblem, out_to_ndarray, check
from pymoo.problems.meta import MetaProblem


def out_to_numpy(out):
    for key in out.keys():
        if type(out[key]) == autograd.numpy.numpy_boxes.ArrayBox:
            out[key] = out[key]._value


def get_deriv(out):
    deriv = []
    if "dF" in out:
        deriv.append("F")
    if "dG" in out:
        deriv.append("G")
    return deriv


def run_and_trace(fun, x, *args, **kwargs):
    start_node = VJPNode.new_root()

    start_box = new_box(x, 0, start_node)
    out = fun(start_box, *args, **kwargs)

    return start_box, out


def calc_jacobian_elem(start, end):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        b = anp.ones(start.shape)
        n = new_box(b, 0, VJPNode.new_root())
        jac = backward_pass(n, end._node)
        return jac._value


def calc_jacobian(start, end):
    # if the end_box is not a box - autograd can not track back
    if not isbox(end):
        return vspace(start.shape).zeros()

    # the final jacobian matrices
    jac = []

    # the backward pass is done for each objective function once
    for j in range(end.shape[1]):
        b = anp.zeros(end.shape)
        b[:, j] = 1
        n = new_box(b, 0, VJPNode.new_root())
        _jac = backward_pass(n, end._node)
        jac.append(_jac)

    jac = anp.stack(jac, axis=1)

    return jac


def autograd_elementwise_eval(problem, X, out, *args, **kwargs):
    deriv = get_deriv(out)

    if len(deriv) > 0:
        out["__autograd__"], _ = run_and_trace(problem._evaluate, X, out, *args, **kwargs)

        for key in deriv:
            val = out.get(key)
            if val is not None:

                if not isinstance(val, anp.ndarray):
                    val = anp.array([val])

                out["d" + key] = calc_jacobian_elem(out["__autograd__"], val)[None, :]

        # make sure all results are pure numpy arrays
        out_to_numpy(out)

    out_to_ndarray(out)
    check(problem, X, out)

    return out


class AutomaticDifferentiation(MetaProblem):

    def __init__(self, problem):
        super().__init__(problem)

    def do(self, X, out, *args, **kwargs):

        problem = self.problem

        if isinstance(problem, ElementwiseProblem):

            # store the original function for evaluation
            func = problem.func_elementwise_eval

            # set a new elementwise function, do the evaluation and revert to default
            problem.func_elementwise_eval = autograd_elementwise_eval
            problem.do(X, out, *args, **kwargs)

            # revert to the old function evaluation
            problem.func_elementwise_eval = func

        else:
            deriv = get_deriv(out)

            if len(deriv) > 0:
                out["__autograd__"], _ = run_and_trace(self.problem.do, X, out, *args, **kwargs)

                for key in deriv:
                    val = out.get(key)

                    if val is not None:

                        if val.ndim == 1:
                            val = val[:, None]

                        jac = calc_jacobian(out["__autograd__"], val)
                        out["d" + key] = jac

                # make sure all results are pure numpy arrays
                out_to_numpy(out)

            else:
                self.problem.do(X, out, *args, **kwargs)



import numpy as np

from pymoo.algorithms.soo.nonconvex.de import DE, DifferentialEvolutionMating
from pymoo.factory import get_problem
from pymoo.optimize import minimize
from pymoo.util.nds.non_dominated_sorting import NonDominatedSorting


# =========================================================================================================
# Implementation
# =========================================================================================================


class SRDE(DE):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _infill(self):
        pop = self.pop

        f = pop.get("F")[:, 0]
        cv = pop.get("CV")[:, 0]

        # print((cv == 0).sum())

        S = np.arange(len(pop))

        feas = (cv == 0)
        n_feas = feas.sum()

        if feas.sum() > 0:
            f_min, cv_min = f[feas].min(), 0.0
        else:
            f_min, cv_min = np.inf, cv[~feas].min()

        if n_feas > 0:

            above_or_equal_fmin = np.where(f >= f_min)[0]
            below_fmin = np.where(f < f_min)[0]

            balance = len(pop) // 2

            # less feasible solutions than desired
            if n_feas < balance:

                n_repl = balance - n_feas

                if len(above_or_equal_fmin) > 0:
                    sort_by_dec_cv = below_fmin[cv[below_fmin].argsort()[::-1]]
                    repl = sort_by_dec_cv[:n_repl]
                    S[repl] = np.random.choice(above_or_equal_fmin, size=len(repl))

            else:

                if len(below_fmin) > 0:

                    n_repl = n_feas - balance

                    sort_by_dec_cv = above_or_equal_fmin[f[above_or_equal_fmin].argsort()[::-1]]
                    repl = sort_by_dec_cv[:n_repl]

                    S[repl] = np.random.choice(below_fmin, size=len(repl))


        mating = DifferentialEvolutionMating()
        infills = mating.do(self.problem, self.pop, self.n_offsprings, S=S, algorithm=self)

        return infills

    def _advance(self, infills=None, **kwargs):
        assert infills is not None, "This algorithms uses the AskAndTell interface thus infills must to be provided."

        pop, off = self.pop, infills
        assert len(pop) == len(off)

        pop_X, pop_f, pop_cv = pop.get("X", "F", "CV")
        pop_f, pop_cv = pop_f[:, 0], pop_cv[:, 0]

        off_X, off_f, off_cv = off.get("X", "F", "CV")
        off_f, off_cv = off_f[:, 0], off_cv[:, 0]

        pop_feas, off_feas = pop_cv == 0, off_cv == 0

        f, cv, feas = np.row_stack([pop_f, off_f]), np.row_stack([pop_cv, off_cv]), np.row_stack([pop_feas, off_feas])

        if feas.sum() > 0:
            f_min, cv_min = f[feas].min(), 0.0
        else:
            f_min, cv_min = np.inf, cv[~feas].min()

        # import matplotlib.pyplot as plt
        #
        # plt.scatter(pop_cv[pop_feas], pop_f[pop_feas], facecolor="none", edgecolors="blue")
        # plt.scatter(pop_cv[~pop_feas], pop_f[~pop_feas], facecolor="none", edgecolors="red")
        #
        # plt.scatter(off_cv[off_feas], off_f[off_feas], facecolor="none", edgecolors="purple", alpha=0.5)
        # plt.scatter(off_cv[~off_feas], off_f[~off_feas], facecolor="none", edgecolors="orange", alpha=0.5)

        for k in range(len(pop)):

            if pop_f[k] >= f_min:

                # if off_f[k] <= pop_f[k]:
                #     self.pop[k] = infills[k]

                if off_cv[k] < pop_cv[k]:
                    self.pop[k] = infills[k]
                elif off_cv[k] == pop_cv[k] and off_f[k] < pop_f[k]:
                    self.pop[k] = infills[k]

            else:

                if off_cv[k] <= pop_cv[k]:
                    self.pop[k] = infills[k]

        # F, CV = self.pop.get("F", "CV")
        # f, cv = F[:, 0], CV[:, 0]
        # plt.scatter(cv, f, facecolor="none", edgecolors="green", s=3, alpha=0.9)
        # plt.show()

    def _advance6(self, infills=None, **kwargs):
        assert infills is not None, "This algorithms uses the AskAndTell interface thus infills must to be provided."

        pop, off = self.pop, infills
        assert len(pop) == len(off)

        N = len(pop)
        pop_X, pop_f, pop_cv = pop.get("X", "F", "CV")
        pop_f, pop_cv = pop_f[:, 0], pop_cv[:, 0]

        off_X, off_f, off_cv = off.get("X", "F", "CV")
        off_f, off_cv = off_f[:, 0], off_cv[:, 0]

        pop_feas, off_feas = pop_cv == 0, off_cv == 0

        f, cv = np.concatenate([pop_f, off_f]), np.concatenate([pop_cv, off_cv])
        nds = NonDominatedSorting().do(np.column_stack([cv, f]), only_non_dominated_front=True)

        pop_cv_nds = set([k for k in nds if k < len(pop)]),
        off_cv_nds = set([k % len(pop) for k in nds if k >= len(pop)])

        import matplotlib.pyplot as plt

        plt.scatter(cv[nds], f[nds], marker="x", color="black", s=30)

        plt.scatter(pop_cv[pop_feas], pop_f[pop_feas], facecolor="none", edgecolors="blue")
        plt.scatter(pop_cv[~pop_feas], pop_f[~pop_feas], facecolor="none", edgecolors="red")

        plt.scatter(off_cv[off_feas], off_f[off_feas], facecolor="none", edgecolors="purple", alpha=0.5)
        plt.scatter(off_cv[~off_feas], off_f[~off_feas], facecolor="none", edgecolors="orange", alpha=0.5)

        hierarch_sort = hierarchical_sort(pop_f, pop_cv)
        fittest = hierarch_sort[0]
        topx = set(hierarch_sort[:10])

        infeas = np.sort(pop_cv[~pop_feas])
        n_infeas = len(pop) // 2

        if len(infeas) == 0:
            eps = np.inf
        else:
            k = min(n_infeas, len(infeas) - 1)
            eps = infeas[k]

        for k in range(len(pop)):

            if k in off_cv_nds:
                if k in pop_cv_nds:
                    if off_cv[k] <= pop_cv[k]:
                        self.pop[k] = infills[k]
                else:
                    if k != fittest:
                        self.pop[k] = infills[k]

            elif k in topx:

                if pop_cv[k] > 0 and off_cv[k] <= pop_cv[k]:
                    self.pop[k] = infills[k]

                else:
                    if off_cv[k] == 0 and off_f[k] <= pop_f[k]:
                        self.pop[k] = infills[k]

            else:

                if off_cv[k] <= eps and pop_cv[k] <= eps:
                    if off_f[k] <= pop_f[k]:
                        self.pop[k] = infills[k]

                elif off_cv[k] <= pop_cv[k]:
                    self.pop[k] = infills[k]

        plt.plot(np.ones(1000) * eps, np.linspace(f.min(), f.max(), 1000), "--", color="black")

        F, CV = self.pop.get("F", "CV")
        f, cv = F[:, 0], CV[:, 0]
        plt.scatter(cv, f, facecolor="none", edgecolors="green", s=3, alpha=0.9)

        plt.xlim(0, cv.max())
        plt.show()

    def _advance3(self, infills=None, **kwargs):
        assert infills is not None, "This algorithms uses the AskAndTell interface thus infills must to be provided."

        pop, off = self.pop, infills
        assert len(pop) == len(off)

        N = len(pop)
        pop_X, pop_f, pop_cv = pop.get("X", "F", "CV")
        pop_f, pop_cv = pop_f[:, 0], pop_cv[:, 0]

        off_X, off_f, off_cv = off.get("X", "F", "CV")
        off_f, off_cv = off_f[:, 0], off_cv[:, 0]

        pop_feas, off_feas = pop_cv == 0, off_cv == 0

        infeas = np.sort(pop_cv[~pop_feas])
        n_infeas = len(pop) // 2
        if len(infeas) == 0:
            eps = np.inf
        elif len(infeas) < n_infeas:
            eps = infeas[n_infeas]

        if pop_feas.sum() > 0:
            f_min, cv_min = pop_f[pop_feas].min(), 0.0
        else:
            f_min, cv_min = np.inf, pop_cv[~pop_feas].min()

        f, cv = np.concatenate([pop_f, off_f]), np.concatenate([pop_cv, off_cv])
        nds = NonDominatedSorting().do(np.column_stack([cv, f]), only_non_dominated_front=True)

        # if pop_feas.sum() > 0:
        #     nds = [k for k in nds if cv[k] < 0.001]

        pop_cv_nds = set([k for k in nds if k < len(pop)]),
        off_cv_nds = set([k % len(pop) for k in nds if k >= len(pop)])

        n_infeas = (~pop_feas).sum()
        n_infeas_goal = int(self.r_infeas * N)

        repl = np.full(N, False)

        for k in range(N):

            if pop_feas[k] and off_feas[k]:
                if off_f[k] <= pop_f[k]:
                    repl[k] = True

            elif pop_feas[k] and ~off_feas[k]:

                if pop_f[k] > f_min:
                    if k in off_cv_nds:
                        repl[k] = True

            elif ~pop_feas[k] and off_feas[k]:

                if k not in pop_cv_nds:
                    repl[k] = True

            elif ~pop_feas[k] and ~off_feas[k]:

                if k in off_cv_nds:

                    if k in pop_cv_nds:

                        if off_cv[k] <= pop_cv[k]:
                            repl[k] = True

                    else:
                        repl[k] = True

                else:
                    if off_cv[k] <= pop_cv[k]:
                        repl[k] = True

        print(repl.sum())

        import matplotlib.pyplot as plt

        plt.scatter(cv[nds], f[nds], marker="x", color="black", s=30)

        F, CV = self.pop.get("F", "CV")
        f, cv = F[:, 0], CV[:, 0]
        feas = cv <= 0

        plt.scatter(cv[feas], f[feas], facecolor="none", edgecolors="blue")
        plt.scatter(cv[~feas], f[~feas], facecolor="none", edgecolors="red")

        F, CV = infills.get("F", "CV")
        f, cv = F[:, 0], CV[:, 0]
        feas = cv <= 0

        plt.scatter(cv[feas], f[feas], facecolor="none", edgecolors="blue", alpha=0.2)
        plt.scatter(cv[~feas], f[~feas], facecolor="none", edgecolors="red", alpha=0.2)

        self.pop[repl] = off[repl]

        F, CV = self.pop.get("F", "CV")
        f, cv = F[:, 0], CV[:, 0]
        plt.scatter(cv, f, facecolor="none", edgecolors="green", s=3, alpha=0.9)

        plt.show()

    def _advance2(self, infills=None, **kwargs):
        assert infills is not None, "This algorithms uses the AskAndTell interface thus infills must to be provided."
        pop, off = self.pop, infills
        N = len(pop)

        if self.t_infeas is None:
            self.t_infeas = np.zeros(N)

        assert len(pop) == len(off)

        pop_f, pop_cv, pop_t = pop.get("F")[:, 0], pop.get("CV")[:, 0], self.n_gen - pop.get("n_gen") - 1
        off_f, off_cv = off.get("F")[:, 0], off.get("CV")[:, 0]
        pop_feas, off_feas = pop_cv == 0, off_cv == 0

        self.t_infeas += 1
        self.t_infeas[pop_feas] = 0

        if pop_feas.sum() > 0:
            f_min, cv_min = pop_f[pop_feas].min(), 0.0
        else:
            f_min, cv_min = np.inf, pop_cv[~pop_feas].min()

        n_infeas = (pop_cv > 0).sum()
        n_goal_infeas = int(N * self.r_infeas)
        n_goal_feas = N - n_goal_infeas

        # if all solutions are currently infeasible
        if pop_feas.sum() == 0:
            for k in range(N):
                if off_cv[k] <= pop_cv[k]:
                    repl[k] = True

        else:

            # S = off_f.argsort()

            for k in range(N):

                if pop_feas[k]:

                    # if both are feasible replace if function value is better
                    if off_feas[k]:
                        if off_f[k] <= pop_f[k]:
                            repl[k] = True

                    # current solution is feasible and offspring is not
                    else:

                        # never replace the current best solution
                        if pop_f[k] > f_min:

                            # only consider a replacement to satisfy the ration of infeasible solutions in pop
                            if n_infeas < n_goal_infeas:
                                if off_f[k] <= pop_f[k]:
                                    repl[k] = True
                                    n_infeas += 1

                # the current solution is infeasible
                else:

                    # we have found a new solution which is feasible and improves the current best
                    if off_feas[k] and off_f[k] <= f_min:
                        repl[k] = True

                    # otherwise we apply our fuzzy logic
                    else:

                        n_feas = N - n_infeas

                        if n_feas < n_goal_feas and off_cv[k] < pop_cv[k]:
                            repl[k] = True
                            n_infeas -= 1

                        # we aim to find a solution which improve f and cv at the same time
                        elif self.t_infeas[k] <= 10:

                            if off_cv[k] < pop_cv[k] and off_f[k] < pop_f[k]:
                                repl[k] = True

                        else:

                            if off_cv[k] < pop_cv[k]:
                                repl[k] = True

        # self.pop[self.indices] = self.survival.do(self.problem, self.pop[self.indices], infills, algorithm=self)

        import matplotlib.pyplot as plt

        F, CV = self.pop.get("F", "CV")
        f, cv = F[:, 0], CV[:, 0]
        feas = cv <= 0

        plt.scatter(cv[feas], f[feas], facecolor="none", edgecolors="blue")
        plt.scatter(cv[~feas], f[~feas], facecolor="none", edgecolors="red")

        F, CV = infills.get("F", "CV")
        f, cv = F[:, 0], CV[:, 0]
        feas = cv <= 0

        plt.scatter(cv[feas], f[feas], facecolor="none", edgecolors="blue", alpha=0.2)
        plt.scatter(cv[~feas], f[~feas], facecolor="none", edgecolors="red", alpha=0.2)

        plt.show()

        self.pop = pop

        # f = self.pop.get("F")[:, 0]
        # cv = self.pop.get("CV")[:, 0]
        #
        # self.pop = self.pop[stochastic_ranking(f, cv, 0.45)]

        # sort the population by fitness to make the selection simpler for mating (not an actual survival, just sorting)
        # self.pop = FitnessSurvival().do(self.problem, self.pop)


if __name__ == "__main__":

    problem = get_problem("g01")

    algorithm = SRDE(pop_size=20, CR=0.3)

    # algorithm = DE(pop_size=100, variant="DE/rand/1/bin", CR=0.5)

    res = minimize(problem,
                   algorithm,
                   ("n_gen", 3000),
                   seed=1,
                   verbose=True)

    print("Best solution found: \nX = %s\nF = %s" % (res.X, res.F))



import numpy as np

from pymoo.problems.meta import MetaProblem


class StochasticTunneling(MetaProblem):

    def __init__(self, problem, fmin, gamma):
        super().__init__(problem)
        self.fmin = fmin
        self.gamma = gamma

    def do(self, x, out, *args, **kwargs):
        super().do(x, out, *args, **kwargs)

        F = out["F"]
        out["__F__"] = F
        # out["F"] = 1 - np.exp(- self.gamma * (F - self.fmin))

        out["F"] = np.tanh(self.gamma * (F - self.fmin))




